#ifndef MOVE_HPP
#define MOVE_HPP

#include <map>
#include <string>

template <typename BoardCoord>
class Move
{
private:
    bool m_legal;   /*!< true if the move is legal */

    int m_color;    /*!< Color from of the player who created the move */

    BoardCoord m_start; /*!< The coordinates where the piece is taken at the start */
    BoardCoord m_dest;  /*!< The coordinates where the piece should be placed at the end of the move */

    typedef std::map<std::string, int> ParameterMap;
    std::map<std::string, int> m_parameters;            /*!< additional parameters about the move (e.g.: roc) */

    /*! \brief Checks if this move if legal using the coordinates system
     *  If the start or the dest coordinates are invalid, the move will be invalidated
     */
    void checkCoordsLegality()
    {
        if ( !m_start.isValid() || !m_dest.isValid() )
        {
            m_legal=false;
        }
    }

public:
    /*! \brief Default constructor producing an illegal move
     */
    Move():m_legal(false) {}

    /*! \brief Basic Move constructor without additional parameters
     *  The constructor will check if the coordonates are valid.
     *  If not, the move will be considered as illegal.
     *  \param color the color of the player moving
     *  \param start the starting point of the move
     *  \param dest the destination of the move
     */
    Move(int color, const BoardCoord& start, const BoardCoord& dest)
        :m_legal(true),m_color(color),m_start(start),m_dest(dest)
    {
        checkCoordsLegality();
    }

    /*! \brief Move constructor with additional parameters
     *  The constructor will check if the coordonates are valid.
     *  If not, the move will be considered as illegal.
     *  \param color the color of the player moving
     *  \param start the starting point of the move
     *  \param dest the destination of the move
     *  \param parameters the additional parameters associated to the move
     */
    Move(int color, const BoardCoord& start, const BoardCoord& dest, const ParameterMap& parameters)
        :m_legal(true),m_color(color),m_start(start),m_dest(dest),m_parameters(parameters)
    {
        checkCoordsLegality();
    }

    /*! \brief Defines a new property
     *  \param name the name to call the property
     *  \param value the value of the property to keep
     */
    void setProperty(const std::string & name, int value)
    {
        // Can overwrite previous value
        m_parameters[name] = value;
    }

    /*! \brief Determines if the move have the property names <a>name</a>
     *  \name the name of the property to check if the move have it
     *  \return true if the property named <a>name</a> does exists
     */
    bool hasProperty(const std::string& name)const
    {
        const ParameterMap::const_iterator itProp = m_parameters.find(name);
        if ( itProp != m_parameters.end() )
        {
            return true;
        }

        return false;
    }

    /*! \brief Returns the value associated to the property named <a>name>/a>
     *  \name the name of the property to get
     *  \return the value of the property
     */
    int getProperty(const std::string& name)const
    {
        const ParameterMap::const_iterator itProp = m_parameters.find(name);
        if ( itProp != m_parameters.end() )
        {
            return itProp->second;
        }

        // Like an invalid value (the user should check it with hasProperty()
        return -1;
    }

    /*! \brief
     *  \return true if the move is legal
     */
    bool isLegal()const { return m_legal; }

    /*! \brief Makes the move illegal
     */
    void invalidate() { m_legal = false; }

    /*! \brief
     *  \return the color of the player initiating this move
     */
    int color()const { return m_color; }

    /*! \brief
     *  \return the start of the move
     */
    const BoardCoord& start()const { return m_start; }

    /*! \brief
     *  \return the destination of the move
     */
    const BoardCoord& dest()const { return m_dest; }
};

/*! \class Move Move.hpp Game/Move.hpp
 *  \brief Class to represent a move in the game
 *
 *  Basically, a move is characterised by a start position
 *  (the position where the piece to move is (<a>start()</a>)) and a destination
 *  (the place where your piece will be (<a>dest()</a>))<br>
 *  Additionnaly to that, the colour completes this structure.<br>
 *  The colour (<a>color()</a>) could be removed, but, we would like to be sure that the player
 *  who generates the move, is the owner of the piece to move.<br>
 *  Since, we want to represent a move for "every" board system possible, the coords
 *  of the origin and destination are reprenseted in a generic way (depending on the board)<br>
 *  Technically, there is not way to know if a move is illegal from itself. So the default
 *  constructor <a>Move()</a> builds an illegal move. Secondly, all moves using invalid coords
 *  is also set to illegal. Moreover, a move this can be set to invalid with <a>invalidate()<a>.
 *  Finally and to be as generic as possible (e.g. in a dungeon & dragons game), you can add
 *  parameters to a move (e.g. bonuses) with <a>setProperty</a>. Such parameters does associate
 *  a value to a name. The existency can be checked using <a>hasProperty</a> and read with <a>getProperty</a><br>
 *  The default constructor <a>Move()</a> create an empty illegal move.<br>
 *  The second constructor create a legal move without any parameters. The move is legal is the coords
 *  used for start and destination position are also legal.<br>
 */

#endif
