#include "ChessRules.hpp"
#include "Rules.hpp"

#include "Game/Boards/ChessBoard.hpp"

#include <iostream>
#include <cassert>
#include <cmath>

const ChessBoard::Coords& ChessRules :: getKingCoords(const std::vector<std::pair<ChessBoard::Coords, Piece> >& pieces, int color)const
{
    for ( std::vector<std::pair<ChessBoard::Coords, Piece> >::const_iterator itPiece = pieces.begin() ; itPiece != pieces.end(); ++ itPiece )
    {
        if ( itPiece->second.type() == PT_KING && itPiece->second.color() == color )
        {
            return itPiece->first;
            break;
        }
    }

    assert(false); // No King ? WTF
    return ChessBoard::InvalidCoords;
}

void ChessRules :: init(ChessBoard& cb)
{
    std::vector<std::pair<ChessBoard::Coords, Piece> > initialPieces;
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(0,0), Piece(PC_WHITE,PT_ROOK)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(1,0), Piece(PC_WHITE,PT_KNIGHT)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(2,0), Piece(PC_WHITE,PT_BISHOP)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(3,0), Piece(PC_WHITE,PT_QUEEN)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(4,0), Piece(PC_WHITE,PT_KING)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(5,0), Piece(PC_WHITE,PT_BISHOP)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(6,0), Piece(PC_WHITE,PT_KNIGHT)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(7,0), Piece(PC_WHITE,PT_ROOK)));
    for ( unsigned int i = 0 ; i < 8 ; i++ )
    {
        initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(i,1), Piece(PC_WHITE,PT_PAWN)));
    }

    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(0,7), Piece(PC_BLACK,PT_ROOK)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(1,7), Piece(PC_BLACK,PT_KNIGHT)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(2,7), Piece(PC_BLACK,PT_BISHOP)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(3,7), Piece(PC_BLACK,PT_QUEEN)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(4,7), Piece(PC_BLACK,PT_KING)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(5,7), Piece(PC_BLACK,PT_BISHOP)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(6,7), Piece(PC_BLACK,PT_KNIGHT)));
    initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(7,7), Piece(PC_BLACK,PT_ROOK)));
    for ( unsigned int i = 0 ; i < 8 ; i++ )
    {
        initialPieces.push_back(std::pair<ChessBoard::Coords, Piece>(ChessBoard::Coords(i,6), Piece(PC_BLACK,PT_PAWN)));
    }

    cb.init(initialPieces);
}

bool ChessRules :: checkMove(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    bool resultChecker = true;
    switch ( cb.get(move.start()).type() )
    {
        case PT_PAWN:
            resultChecker = checkPawnMove(cb,move);
            break;
        case PT_ROOK:
            resultChecker = checkRookMove(cb,move);
            break;
        case PT_KNIGHT:
            resultChecker = checkKnightMove(cb,move);
            break;
        case PT_BISHOP:
            resultChecker = checkBishopMove(cb,move);
            break;
        case PT_QUEEN:
            resultChecker = checkQueenMove(cb,move);
            break;
        case PT_KING:
            resultChecker = checkKingMove(cb,move);
            break;

        case RulesUtils::PT_NONE:
        default:
            return false;
            break;
    }

    return resultChecker;
}

bool ChessRules::checkCheck(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    const Piece& startPiece = cb.get(move.start());

    std::vector<std::pair<ChessBoard::Coords, Piece> > pieces = cb.getPieces();

    ChessBoard::Coords kingCoords;

    // Search for the king
    // The king is at the actual position if the move does not move him
    // But in this case, we have to check with the destination
    if ( startPiece.type() == PT_KING)
    {
        // Since we are checking the check when the king will be at new position
        kingCoords = move.dest();
    }
    else // Otherwise, we have to look after him
    {
        kingCoords = getKingCoords(pieces,move.color());
    }

    // For each ennemy, we check if one of these piece can reache the king
    ChessBoard simulationCB(cb);
    simulationCB.move(move);

    pieces = simulationCB.getPieces();
    if ( isCheck(simulationCB,pieces,kingCoords,move.color()) )
    {
        return true;
    }

    return false;
}

bool ChessRules::check(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    ChessBoard::Coords startCoords = move.start();
    const Piece& startPiece = cb.get(startCoords);
    const Piece& endPiece = cb.get(move.dest());

    // First checks ->
    // If the player tries to move a piece not handled by him
    //  OR
    // If there is no piece at start position (no piece to move)
    if ( startPiece.color() != move.color() || !RulesUtils::isAPiece(startPiece) )
    {
        std::cout << "Fail since the move does not have the color of the piece, or the start position is pointing not a piece" << std::endl;
        return false;
    }

    // If there is another piece at the destination, and that this one is from the same player
    // we are sure that the movement is impossible
    if ( RulesUtils::isAPiece(endPiece) && (startPiece.color() == endPiece.color()) )
    {
        return false;
    }

    // If it is not a move
    if ( (move.start().column == move.dest().column) &&
         (move.start().line == move.dest().line) )
    {
        return false;
    }


    bool resultChecker = checkMove(cb,move);
    // Last check is to know if we will be chess
    if ( resultChecker )
    {
        if ( checkCheck(cb,move) )
        {
            resultChecker = false;
        }
    }

    return resultChecker;
}

bool ChessRules::checkPawnMove(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    const Piece& startPiece = cb.get(move.start());
    int ahead = 0;
    bool priseEnPassant = false;

    int distanceY = move.dest().line - move.start().line;

    switch ( startPiece.color() )
    {
    case PC_WHITE:
        ahead = 1;
        if ( distanceY == 2 )
        {
            if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords(move.start().column,move.start().line+ahead))) ||
                 RulesUtils::isAPiece(cb.get(move.dest())) )
            {
                return false;
            }

            if ( move.start().line == 1 )
            {
                // Cheating to make the following test working :)
                distanceY = 1;
            }
        }

        if ( move.start().line == 4 )
        {
            priseEnPassant = true;
        }

        break;
    case PC_BLACK:
        ahead = -1;
        if ( distanceY == -2 )
        {
            if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords(move.start().column,move.start().line+ahead))) ||
                 RulesUtils::isAPiece(cb.get(move.dest())) )
            {
                return false;
            }

            if ( move.start().line == 6 )
            {
                // Cheating to make the following test working :)
                distanceY = -1;
            }
        }

        if ( move.start().line == 3 )
        {
            priseEnPassant = true;
        }
        break;
    default:
        // The piece is not chess piece (because of the color)
        assert(false);
        return false;
        break;
    }

    // Check if the Pawn is moved one case ahead
    if (distanceY != ahead)
    {
        return false;
    }

    // Promotion
    if ( move.dest().line == 7 || move.dest().line == 0 )
    {
        if ( move.hasProperty("promotion") )
        {
            char promotion = move.getProperty("promotion");
            switch(promotion)
            {
                // We update the property to have directly the ID (usable by ChessBoard)
                // This rely on the fact that setProperty overwrite
                case 'q':
                    move.setProperty("promotionID",PT_QUEEN);
                    return true;
                    break;
                case 'b':
                    move.setProperty("promotionID",PT_BISHOP);
                    return true;
                    break;
                case 'r':
                    move.setProperty("promotionID",PT_ROOK);
                    return true;
                    break;
                case 'n':
                    move.setProperty("promotionID",PT_KNIGHT);
                    return true;
                    break;
                default:
                    std::cout << "Invalid promotion" << std::endl;
                    return false;
            }
        }
        else
        {
            return false;
        }
    }

    // On the same column, so we are moving straight ahead
    if ( move.dest().column == move.start().column)
    {
        if ( !RulesUtils::isAPiece(cb.get(move.dest())) )
        {
            return true;
        }

        // There is a piece, and we can't move because of this one
        return false;
    }
    else
    {
        // We can only move diagonaly
        if ( std::abs(static_cast<float>(move.dest().column - move.start().column)) != 1 )
        {
            return false;
        }

        if ( RulesUtils::isAPiece(cb.get(move.dest())) || // if the dest is a piece (attack)
             (RulesUtils::isAPiece(cb.get(ChessBoard::Coords(move.dest().column,move.start().line))) && priseEnPassant) ) //if the case next to him is a piece (prise en passant)
        {
            return true;
        }
        return false;
    }
}

bool ChessRules::checkRookMove(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    // First, check if the are moving only following a line
    int distanceX = move.dest().column - move.start().column;
    int distanceY = move.dest().line - move.start().line;

    if ( ((distanceX != 0) && (distanceY == 0)) || (distanceX == 0 && distanceY != 0) )
    {
        // We have to check if there is something between the start and the dest
        if ( distanceX != 0 )
        {
            if ( distanceX < 0 )
            {
                for ( int i = -1 ; i > distanceX ; --i )
                {
                    if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords(move.start().column+i,move.start().line))) )
                    {
                        return false;
                    }
                }
            }
            else
            {
                for ( int i = 1 ; i < distanceX ; ++i )
                {
                    if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords(move.start().column+i,move.start().line))) )
                    {
                        return false;
                    }
                }
            }
        }
        if ( distanceY != 0 )
        {
            if ( distanceY < 0 )
            {
                for ( int i = -1 ; i > distanceY ; --i )
                {
                    if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords(move.start().column,move.start().line+i))) )
                    {
                        return false;
                    }
                }
            }
            else
            {
                for ( int i = 1 ; i < distanceY ; ++i )
                {
                    if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords(move.start().column,move.start().line+i))) )
                    {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    return false;
}

bool ChessRules::checkKnightMove(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    (void)cb;

    float distanceX = std::abs(static_cast<float>(move.dest().column - move.start().column));
    float distanceY = std::abs(static_cast<float>(move.dest().line - move.start().line));

    // Knight is quite ease since it does not have to check if there is something between start and dest
    // Moreover, is just a 2 / 1 move
    if ( distanceX >= 1 && distanceX <= 2 &&
         distanceY >= 1 && distanceY <= 2 )
    {
        // Maybe this test can completely replace the previous one
        if ( (distanceX == 2 && distanceY == 1) || (distanceX == 1 && distanceY == 2) )
        {
            return true;
        }

        return false;
    }

    return false;
}

bool ChessRules::checkBishopMove(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    int distanceX = move.dest().column - move.start().column;
    int distanceY = move.dest().line - move.start().line;

    // Check for diagonals
    if ( std::abs(static_cast<float>(distanceX)) == std::abs(static_cast<float>(distanceY)) )
    {
        // We have to go through all cases between here and destination.
        int stepX = 0;
        int stepY = 0;

        if ( distanceX < 0)
        {
            stepX = -1;
        }
        else
        {
            stepX = 1;
        }

        if ( distanceY < 0 )
        {
            stepY = -1;
        }
        else
        {
            stepY = 1;
        }

        for ( int i = 1 ; i < std::abs(static_cast<float>(distanceX)) /*OR distanceY since we are the same*/ ; i++ )
        {
            // We have a piece on the way
            if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords( move.start().column+i*stepX,move.start().line+i*stepY))) )
            {
                return false;
            }
        }

        return true;
    }

    return false;
}

bool ChessRules::checkQueenMove(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    bool isPossible = false;

    // Queen is so cheated :D
    isPossible |= checkRookMove(cb,move);
    isPossible |= checkBishopMove(cb,move);

    return isPossible;
}

bool ChessRules::checkKingMove(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    int distanceX = move.dest().column - move.start().column;
    int distanceY = move.dest().line - move.start().line;

    if ( std::abs(static_cast<float>(distanceX)) <= 1 && std::abs(static_cast<float>(distanceY)) <= 1 )
    {
        return true;
    }
    else if ( std::abs(static_cast<float>(distanceX)) == 2 && distanceY == 0 ) // Rock
    {
        return checkCastlingMove(cb,move);
    }

    return false;
}

bool ChessRules :: checkCastlingMove(const ChessBoard& cb, Move<ChessBoard::Coords>& move)const
{
    int distanceX = move.dest().column - move.start().column;
    const Piece& startPiece = cb.get(move.start());

    if ( startPiece.getProperty("hasMoved") == false ) // If the piece never moved
    {
        int line = 0;
        if ( startPiece.color() == PC_WHITE )
        {
            line = 0;
        }
        else
        {
            line = 7;
        }

        if ( move.start().column == 4 && move.start().line == line ) // If the king is really at the correct position
        {
            Piece rookPiece;
            int coords[2] = {0, 0};
            if ( distanceX < 0 )
            {
                coords[0] = 7;
                rookPiece = cb.get(coords);
                // Check empty case between king and rook
                if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords(5,line))) || RulesUtils::isAPiece(cb.get(ChessBoard::Coords(6,line))) )
                {
                    return false;
                }
            }
            else
            {
                coords[0] = 0;
                rookPiece = cb.get(coords);
                if ( RulesUtils::isAPiece(cb.get(ChessBoard::Coords(1,line))) ||
                     RulesUtils::isAPiece(cb.get(ChessBoard::Coords(2,line))) ||
                     RulesUtils::isAPiece(cb.get(ChessBoard::Coords(3,line))) )
                {
                    return false;
                }
            }

            if ( rookPiece.type() == PT_ROOK ) // If the rook is at the place correct place
            {
                if ( rookPiece.getProperty("hasMoved") == false ) // If the rook moved once ?
                {
                    std::vector<std::pair<ChessBoard::Coords, Piece> > pieces = cb.getPieces();
                    ChessBoard simulationCB(cb);

                    if ( distanceX > 0 )
                    {
                        simulationCB.move(Move<ChessBoard::Coords>(startPiece.color(),move.start(), ChessBoard::Coords(3,line)));
                    }
                    else
                    {
                        simulationCB.move(Move<ChessBoard::Coords>(startPiece.color(),move.start(), ChessBoard::Coords(5,line)));
                    }

                    // For each ennemy, we check if one of these piece can reache the king
                    for ( std::vector<std::pair<ChessBoard::Coords, Piece> >::const_iterator itPiece = pieces.begin() ; itPiece != pieces.end(); ++ itPiece )
                    {
                        if ( itPiece->second.color() != move.color() )
                        {
                            Move<ChessBoard::Coords> chessMove(itPiece->second.color(),itPiece->first,move.start());

                            if ( checkMove(simulationCB,chessMove) )
                            {
                                // We can attack the king, we are in chess
                                return false;
                            }
                        }
                    }

                    ChessBoard::Coords kingCoords;
                    if ( distanceX > 0 )
                    {
                        kingCoords = ChessBoard::Coords(3,line);
                        simulationCB.move(Move<ChessBoard::Coords>(startPiece.color(),kingCoords, ChessBoard::Coords(2,line)));
                    }
                    else
                    {
                        simulationCB.move(Move<ChessBoard::Coords>(startPiece.color(),kingCoords, ChessBoard::Coords(6,line)));
                        kingCoords = ChessBoard::Coords(5,line);
                    }


                    // For each ennemy, we check if one of these piece can reache the king
                    for ( std::vector<std::pair<ChessBoard::Coords, Piece> >::const_iterator itPiece = pieces.begin() ; itPiece != pieces.end(); ++ itPiece )
                    {
                        if ( itPiece->second.color() != move.color() )
                        {
                            Move<ChessBoard::Coords> chessMove(itPiece->second.color(),itPiece->first,kingCoords);

                            if ( checkMove(simulationCB,chessMove) )
                            {
                                // We can attack the king, we are in chess
                                return false;
                            }
                        }
                    }
                }
            }
        }
    }

    return true;
}


bool ChessRules :: isCheck(const ChessBoard& cb, const std::vector<std::pair<ChessBoard::Coords, Piece> >& pieces, const ChessBoard::Coords& kingCoords, const int color)const
{
    for ( std::vector<std::pair<ChessBoard::Coords, Piece> >::const_iterator itPiece = pieces.begin() ; itPiece != pieces.end(); ++ itPiece )
    {
        if ( itPiece->second.color() != color )
        {
            Move<ChessBoard::Coords> chessMove(itPiece->second.color(),itPiece->first,kingCoords);
            if ( checkMove(cb,chessMove) )
            {
                // We can attack the king, we are in check
                return true;
            }
        }
    }

    return false;
}

bool ChessRules :: checkVictory(const ChessBoard& cb, const int color)const
{
    // We get the color of the opponent, since victory is based on the oppenent defeat
    int opponentColor = 1 - color;

    std::vector<std::pair<ChessBoard::Coords, Piece> > pieces = cb.getPieces();
    ChessBoard::Coords kingCoords = getKingCoords(pieces,opponentColor);

    if ( !isCheck(cb,pieces,kingCoords,opponentColor) )
    {
        return false;
    }

    const int kingMoves[][2] =
    {
        { -1, -1 } , { 0, -1 } , { 1, -1 },
        { -1, 0 }  ,             { 1, 0 },
        { -1, 1 } , { 0, 1 } , { 1, 1 }
    };

    // We check if a move of the king would resolve the check
    for ( unsigned int i = 0 ; i < sizeof(kingMoves)/(2*sizeof(int)) ; i++ )
    {
        ChessBoard simulationCB(cb);
        ChessBoard::Coords destination(kingCoords.column+kingMoves[i][0],kingCoords.line+kingMoves[i][1]);
        Piece destinationPiece = cb.get(destination);

        if ( !destination.isValid() )
        {
            continue;
        }

        // We can't move since there is a piece of the same color
        if ( RulesUtils::isAPiece(destinationPiece) && destinationPiece.color() == opponentColor )
        {
            continue;
        }

        simulationCB.move(Move<ChessBoard::Coords>(opponentColor,kingCoords,destination));
        if ( !isCheck(simulationCB,pieces,destination,opponentColor) )
        {
            return false;
        }
    }
    // Now, we have to try to check all this player movement, to see if we can avoid the checkmate
    // This is the naive way to do it, I am just checking every possible solution :(
    for ( std::vector<std::pair<ChessBoard::Coords, Piece> >::const_iterator itPiece = pieces.begin() ; itPiece != pieces.end(); ++ itPiece )
    {
        if ( itPiece->second.color() == opponentColor && itPiece->second.type() != PT_KING ) // We do not retest the king
        {
            for ( int column = 0 ; column < cb.getDimensionLength(0) ; column++ )
            {
                for ( int line = 0 ; line < cb.getDimensionLength(1) ; line++ )
                {
                    int destination[2] = { column , line };
                    Move<ChessBoard::Coords> protectingMove(opponentColor,itPiece->first,destination);
                    if ( this->check(cb,protectingMove) )
                    {
                        ChessBoard simulationCB(cb);
                        simulationCB.move(protectingMove);
                        if ( !isCheck(simulationCB,pieces,kingCoords,opponentColor) )
                        {
                            // No longer is check (so this move can protect the player)
                            return false;
                        }
                    }
                }
            }

        }
    }

    // No solution, this is a checkmate
    return true;
}



bool operator!=(const ChessBoard::Coords& c1, const ChessBoard::Coords& c2)
{
    return c1.column != c2.column || c1.line != c2.line;
}

bool ChessRules::checkDraw(const History<Move<ChessBoard::Coords> >& history)const
{
    // Checks about the rule of 50
    if ( m_nbTurnsPat >= 50 )
    {
        return true;
    }

    // Checks about three time the same move
    int historySize = history.size();
    std::vector<std::pair<Move<ChessBoard::Coords>,Move<ChessBoard::Coords> > > moves;
    if ( historySize > 3 )
    {
        for ( int i = historySize - 1 ; i > historySize - 4 ; i-- )
        {
            moves.push_back(history.get(i));
        }
    }
    else
    {
        return false;
    }

    bool patWhite = true;
    for ( unsigned int i = 1 ; i < moves.size() ; i++ )
    {
        if ( moves[i-1].first.start() != moves[i].first.dest() ||
             moves[i-1].first.dest() != moves[i].first.start() )
        {
            patWhite = false;
            break;
        }
    }

    bool patDark = true;
    for ( unsigned int i = 1 ; i < moves.size() ; i++ )
    {
        if ( moves[i-1].second.start() != moves[i].second.dest() ||
             moves[i-1].second.dest() != moves[i].second.start() )
        {
            patDark = false;
            break;
        }
    }

    return patDark || patWhite;
}

void ChessRules::move(const ChessBoard& cb, const Move<ChessBoard::Coords>& move)
{
    const Piece& startPiece = cb.get(move.start());
    const Piece& endPiece = cb.get(move.dest());

    if ( !move.isLegal() )
    {
        return;
    }

    // Update the Pat counter (50 turns)
    m_nbTurnsPat += 0.5f; // Once it will reach 50, we will stop

    // Pawn move, we can reset the counter
    if ( startPiece.type() == PT_PAWN )
    {
        m_nbTurnsPat = 0;
    }

    // If the destination is a piece,
    if ( RulesUtils::isAPiece(endPiece) )
    {
        m_nbTurnsPat = 0;
    }
}
