#include "Qt2DDisplay.hpp"

#include <QGraphicsRectItem>
#include <QResizeEvent>
#include <QDebug>

#include "Game/Rules/Rules.hpp"
#include "Game/Rules/ChessRules.hpp"
#include "Game/Boards/ChessBoard.hpp"

Qt2DDisplay::Qt2DDisplay(QWidget* parent)
    :QGraphicsView(parent)
{
    this->loadPieces();
    this->setScene(&m_scene);
    // this->setWidget(m_view);

    connect(this,SIGNAL(gameUpdateSignal()),this,SLOT(updateGame()));
    connect(this,SIGNAL(pieceMoved(CMove)),this,SLOT(movePiece(CMove)));

    //typedef QString CustomString;
    qRegisterMetaType<CMove>("CMove");
    //qRegisterMetaType("CMove");

    // setAcceptedMouseButtons(Qt::RightButton);
}

void Qt2DDisplay::loadPieces()
{
    m_piecesPixmap[1] = std::pair<QString,QString>(":/pieces/wpawn.svg",":/pieces/bpawn.svg");
    m_piecesPixmap[2] = std::pair<QString,QString>(":/pieces/wrook.svg",":/pieces/brook.svg");
    m_piecesPixmap[3] = std::pair<QString,QString>(":/pieces/wknight.svg",":/pieces/bknight.svg");
    m_piecesPixmap[4] = std::pair<QString,QString>(":/pieces/wbishop.svg",":/pieces/bbishop.svg");
    m_piecesPixmap[5] = std::pair<QString,QString>(":/pieces/wqueen.svg",":/pieces/bqueen.svg");
    m_piecesPixmap[6] = std::pair<QString,QString>(":/pieces/wking.svg",":/pieces/bking.svg");
/*
    // Check if all loading are done
    for ( std::map<std::pair<QPixmap, QPixmap> >::const_iterotor itPair = m_piecesPixmap.begin() ; itPair != m_piecesPixmap.end() ; ++itPair)
    {

    }*/
}

Qt2DDisplay::~Qt2DDisplay()
{

}

void Qt2DDisplay::createBoard()
{
    for ( int column = 0 ; column < m_pBoard->getDimensionLength(0) ; column ++ )
    {
        for ( int line = 0 ; line < m_pBoard->getDimensionLength(1) ; line++ )
        {
            QGraphicsRectItem* pRect = m_scene.addRect(QRect(column*m_caseWidth,line*m_caseHeight,m_caseWidth,m_caseHeight));
            if ( ((column+line) % 2) == 0 )
            {
                pRect->setBrush(QColor(255,206,158));
            }
            else
            {
                pRect->setBrush(QColor(209,139,71));
            }
        }
    }
}

void Qt2DDisplay::clearScreen()
{
    clearPieces();
    m_scene.clear();
}

void Qt2DDisplay::clearPieces()
{
    for ( std::vector<Qt2DPiece*>::const_iterator itItem = m_pieces.begin() ; itItem != m_pieces.end() ; ++itItem )
    {
        m_scene.removeItem(*itItem);
        delete *itItem;
    }

    m_pieces.clear();
}

void Qt2DDisplay::updateGame()
{
    this->clearPieces();
    displayBoard();
}

void Qt2DDisplay::resizeEvent(QResizeEvent *event)
{
    m_caseWidth = event->size().width() / m_pBoard->getDimensionLength(0);
    m_caseHeight = event->size().height() / m_pBoard->getDimensionLength(1);

    clearScreen();
    createBoard();
    displayBoard();
}

void Qt2DDisplay::movePiece(Move<ChessBoard::Coords> move)
{
    int pieceX = m_caseWidth * move.dest().column;
    int pieceY = m_caseHeight * move.dest().line;

    // Move animation for the piece
    QGraphicsItem* pItem = m_scene.itemAt(move.start().column*m_caseWidth + m_caseWidth/2, move.start().line*m_caseHeight+m_caseHeight/2);
    if ( pItem )
    {
        float scaleX = m_caseWidth/pItem->boundingRect().width();
        float scaleY = m_caseHeight/pItem->boundingRect().height();
        float scale = std::min(scaleX,scaleY);

        Qt2DPiece* pPiece = dynamic_cast<Qt2DPiece*>(pItem);
        if ( pPiece )
        {
            QPropertyAnimation* animation = new QPropertyAnimation(pPiece, "pos",this);
            animation->setDuration(1500);
            animation->setStartValue(pPiece->pos());
            animation->setEndValue(QPointF(pieceX + (m_caseWidth-(scale*pItem->boundingRect().width()))/2.0f,
                                          pieceY + (m_caseHeight-(scale*pItem->boundingRect().height()))/2.0f));
            animation->setEasingCurve(QEasingCurve::InOutQuad);
            animation->start();

            // Cause a complete redraw of the game at end
            connect(animation,SIGNAL(finished()),this,SLOT(updateGame()));

            // Disapearing animation for the dest piece (if one)
            QGraphicsItem* pItemDest = m_scene.itemAt(move.dest().column*m_caseWidth + m_caseWidth/2, move.dest().line*m_caseHeight+m_caseHeight/2);
            if ( pItemDest )
            {
                Qt2DPiece* pPiece = dynamic_cast<Qt2DPiece*>(pItemDest);
                if ( pPiece )
                {
                    QPropertyAnimation* animation = new QPropertyAnimation(pPiece, "opacity",this);
                    animation->setDuration(1500);
                    animation->setStartValue(pPiece->opacity());
                    animation->setEndValue(0);
                    animation->setEasingCurve(QEasingCurve::InOutQuad);
                    animation->start();
                }
            }
        }
        else
        {
            updateGame();
        }
    }
}

void Qt2DDisplay::setBoard(const ChessBoard *pBoard)
{
    assert(pBoard);

    m_pBoard = pBoard;

    m_caseWidth = this->width() / m_pBoard->getDimensionLength(0);
    m_caseHeight = this->height() / m_pBoard->getDimensionLength(1);

    this->resize(600,400);
}

void Qt2DDisplay::collectMove(QPointF start, QPointF end, int color)
{
    int columnStart = start.x()/m_caseWidth;
    int lineStart = start.y()/m_caseHeight;
    int columnEnd = end.x()/m_caseWidth;
    int lineEnd = end.y()/m_caseHeight;

    if ( columnStart != columnEnd || lineStart != lineEnd )
    {
        Move<ChessBoard::Coords> m(color,
                                   ChessBoard::Coords(columnStart,lineStart),
                                   ChessBoard::Coords(columnEnd,lineEnd));

        Piece p = m_pBoard->get(m.start());
        if ( p.isValid() && p.type() == ChessRules::PT_PAWN )
        {
            bool promotion = false;
            if ( color == 0 && m.dest().line == 7 )
            {
                promotion = true;
            }
            else if ( color == 1 && m.dest().line == 0 )
            {
                promotion = true;
            }

            if ( promotion )
            {
                PromotionDialog pm(color,this);
                pm.exec();

                m.setProperty("promotion",pm.getChoice());
            }
        }

        emit movePlayed(m);
    }
    else
    {
        // If start and end are the same, we forget it
        Qt2DPiece* pPiece = dynamic_cast<Qt2DPiece*>(QObject::sender());
        if ( pPiece )
        {
            pPiece->reset();
        }
        else
        {
            // Woot !
            assert(false);
        }
    }
}

void Qt2DDisplay::notify(const Move<ChessBoard::Coords> &move)
{
    if ( move.isLegal() )
    {
        emit pieceMoved(move);
    }
    else
    {
        QGraphicsItem* pItemStart = m_scene.itemAt(move.start().column*m_caseWidth + m_caseWidth/2, move.start().line*m_caseHeight+m_caseHeight/2);
        QGraphicsItem* pItemDest = m_scene.itemAt(move.dest().column*m_caseWidth + m_caseWidth/2, move.dest().line*m_caseHeight+m_caseHeight/2);

        if ( pItemStart )
        {
            Qt2DPiece* pPiece = dynamic_cast<Qt2DPiece*>(pItemStart);
            if ( pPiece )
            {
                pPiece->reset();
            }
        }

        if ( pItemDest )
        {
            Qt2DPiece* pPiece = dynamic_cast<Qt2DPiece*>(pItemDest);
            if ( pPiece )
            {
                pPiece->reset();
            }
        }
    }
}

void Qt2DDisplay::displayHelper(QPointF pos, int color)
{
    int columnStart = pos.x()/m_caseWidth;
    int lineStart = pos.y()/m_caseHeight;
    ChessBoard::Coords start(columnStart,lineStart);

    for ( int column = 0 ; column < m_pBoard->getDimensionLength(0) ; column ++ )
    {
        for ( int line = 0 ; line < m_pBoard->getDimensionLength(1) ; line++ )
        {
            ChessBoard::Coords dest(column,line);
            Move<ChessBoard::Coords> m(color,
                                       start,
                                       dest);

            if ( m_pRules->check(*m_pBoard,m) == true )
            {
                QList<QGraphicsItem*> items = m_scene.items(QPointF(column * m_caseWidth + m_caseWidth/2,
                                                                    line * m_caseHeight + m_caseHeight/2));

                QGraphicsRectItem* pRect = m_scene.addRect(QRect(column*m_caseWidth,line*m_caseHeight,m_caseWidth,m_caseHeight));
                if ( RulesUtils::isAPiece(m_pBoard->get(dest)) ) // Case to eat in blue
                {
                    pRect->setBrush(QColor(134,113,200,128));
                }
                else
                {
                    pRect->setBrush(QColor(149,200,112,128));
                }
                m_helpers.push_back(pRect);
            }
        }
    }

    const std::vector<std::pair<ChessBoard::Coords, Piece> > pieces = m_pBoard->getPieces();
    for ( std::vector<std::pair<ChessBoard::Coords, Piece> >::const_iterator itPiece = pieces.begin() ; itPiece != pieces.end(); ++ itPiece )
    {
        if ( itPiece->second.color() != color )
        {
            Move<ChessBoard::Coords> m(itPiece->second.color(),
                                       itPiece->first,
                                       start); // It means that the destination is actual selection

            if ( m_pRules->check(*m_pBoard,m) )
            {
                QGraphicsRectItem* pRect = m_scene.addRect(QRect(itPiece->first.column*m_caseWidth,itPiece->first.line*m_caseHeight,m_caseWidth,m_caseHeight));
                pRect->setBrush(QColor(255,89,95,128));
                m_helpers.push_back(pRect);
            }
        }
    }
}

void Qt2DDisplay::stopHelper()
{
    for ( std::vector<QGraphicsRectItem*>::iterator itPItem = m_helpers.begin() ; itPItem != m_helpers.end() ; ++itPItem )
    {
        m_scene.removeItem(*itPItem);
        delete *itPItem;
    }
    m_helpers.clear();
}
